"""Classes representing defects."""
from __future__ import annotations

import collections
import logging
from abc import ABCMeta, abstractmethod, abstractproperty
from enum import Enum
from typing import Dict

import numpy as np
from monty.json import MSONable
from pymatgen.analysis.structure_matcher import ElementComparator, StructureMatcher
from pymatgen.core import Element, PeriodicSite, Species, Structure
from pymatgen.core.periodic_table import DummySpecies
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.symmetry.structure import SymmetrizedStructure

from pymatgen.analysis.defects.supercells import get_sc_fromstruct

from .utils import get_plane_spacing

# TODO Possible redesign idea: ``DefectSite`` class defined with a defect object.
# This makes some of the accounting logic a bit harder since we will probably
# just have one concrete ``Defect`` class so you can write custom multiplicity functions
# but it makes the implementation of defect complexes trivial.
# i.e. each defect will be defined by a structure and a collection of ``DefectSite``s

__author__ = "Jimmy-Xuan Shen"
__copyright__ = "Copyright 2022, The Materials Project"
__maintainer__ = "Jimmy-Xuan Shen @jmmshn"
__date__ = "Mar 15, 2022"

_logger = logging.getLogger(__name__)


class DefectType(Enum):
    """Defect type, for sorting purposes."""

    Vacancy = 0
    Substitution = 1
    Interstitial = 2
    Other = 3


class Defect(MSONable, metaclass=ABCMeta):
    """Abstract class for a single point defect."""

    def __init__(
        self,
        structure: Structure,
        site: PeriodicSite,
        multiplicity: int | None = None,
        oxi_state: float | None = None,
        equivalent_sites: list[PeriodicSite] | None = None,
        symprec: float = 0.01,
        angle_tolerance: float = 5,
        user_charges: list[int] | None = None,
    ) -> None:
        """Initialize a defect object.

        Args:
            structure: The structure of the defect.
            site: The site
            multiplicity: The multiplicity of the defect.
            oxi_state: The oxidation state of the defect, if not specified,
                this will be determined automatically.
            equivalent_sites: A list of equivalent sites for the defect in the structure.
            symprec: Tolerance for symmetry finding.
            angle_tolerance: Angle tolerance for symmetry finding.
            user_charges: User specified charge states. If specified,
                ``get_charge_states`` will return this list. If ``None`` or empty list
                the charge states will be determined automatically.
        """
        self.structure = structure
        self.site = site
        self.symprec = symprec
        self.angle_tolerance = angle_tolerance
        self.multiplicity = (
            multiplicity if multiplicity is not None else self.get_multiplicity()
        )
        self.equivalent_sites = equivalent_sites if equivalent_sites is not None else []
        self.user_charges = user_charges if user_charges else []
        if oxi_state is None:
            # Try to use the reduced cell first since oxidation state assignment
            # scales poorly with systems size.
            try:
                self.structure.add_oxidation_state_by_guess(max_sites=-1)
                # check oxi_states assigned and not all zero
                if all(specie.oxi_state == 0 for specie in self.structure.species):
                    self.structure.add_oxidation_state_by_guess()
            except:  # pragma: no cover
                self.structure.add_oxidation_state_by_guess()
            self.oxi_state = self._guess_oxi_state()
        else:
            self.oxi_state = oxi_state

    @abstractmethod
    def get_multiplicity(self) -> int:
        """Get the multiplicity of the defect.

        Returns:
            int: The multiplicity of the defect.
        """

    @abstractmethod
    def _guess_oxi_state(self) -> float:
        """Best guess for the oxidation state of the defect.

        Returns:
            float: The oxidation state of the defect.
        """

    @abstractmethod
    def __repr__(self) -> str:
        """Representation of the defect."""

    @abstractproperty
    def name(self) -> str:
        """Name of the defect."""

    @abstractproperty
    def defect_structure(self) -> Structure:
        """Get the unit-cell structure representing the defect."""

    @abstractproperty
    def element_changes(self) -> Dict[Element, int]:
        """Get the species changes of the defect.

        Returns:
            Dict[Element, int]: The species changes of the defect.
        """

    @property
    def centered_defect_structure(self) -> Structure:
        """Get a defect structure that is centered around the site.

        Move all the sites in the structure so that they
        are in the periodic image closest to the defect site.
        """
        return center_structure(self.defect_structure, self.site.frac_coords)

    def get_charge_states(self, padding: int = 1) -> list[int]:
        """Potential charge states for a given oxidation state.

        If user charges are specified, these will be returned.
        Otherwise, the charge states will be determined automatically based
        on the oxidation state with a padding on either sites of 0 and the
        oxidation state value.

        Args:
            padding: The number of charge states on the on either side of
                0 and the oxidation state.

        Returns:
            list of possible charge states
        """
        if self.user_charges:
            return self.user_charges

        if isinstance(self.oxi_state, int) or self.oxi_state.is_integer():
            oxi_state = int(self.oxi_state)
        else:
            raise ValueError("Oxidation state must be an integer")

        if oxi_state >= 0:
            charges = [*range(-padding, oxi_state + padding + 1)]
        else:
            charges = [*range(oxi_state - padding, padding + 1)]

        return charges

    def get_supercell_structure(
        self,
        sc_mat: np.ndarray | None = None,
        defect_structure: Structure | None = None,
        dummy_species: str | None = None,
        min_atoms: int = 80,
        max_atoms: int = 240,
        min_length: float = 10.0,
        force_diagonal: bool = False,
        relax_radius: float | str | None = None,
        perturb: float | None = None,
        target_frac_coords: np.ndarray | None = None,
        return_site: bool = False,
    ) -> Structure:
        """Generate the supercell for a defect.

        If the bulk structure (provided by `Defect.structure`) and defect structures
        have oxidation state information, then the supercell will be decorated with
        oxidation states.  Otherwise, the supercell structure will not have any oxidation
        state information.  This also allows for oxidation state decoration of different
        defect charge states.

        .. code-block:: python
            defect_struct = defect.defect_structure.copy()
            defect_struct.add_oxidation_state_by_guess(target_charge=2)

        .. note::
            Since any algorithm for decorating the oxidation states will be combinatorial,
            They can be very slow for large supercells. If you want to decorate the structure
            with oxidation states, you have to first decorate the smaller unit cell defect
            structure then implant it in the supercell.

        Args:
            sc_mat: supercell matrix if None, the supercell will be determined by `CubicSupercellAnalyzer`.
            defect_structure: Alternative defect structure to use for generating the supercell.
                You might want to use this if you want to decorate the oxidation states of the
                defect structure using a custom method.
            dummy_species: Dummy species to highlight the defect position (for visualizing vacancies).
            max_atoms: Maximum number of atoms allowed in the supercell.
            min_atoms: Minimum number of atoms allowed in the supercell.
            min_length: Minimum length of the smallest supercell lattice vector.
            force_diagonal: If True, return a transformation with a diagonal transformation matrix.
            relax_radius: Relax the supercell atoms to a sphere of this radius around the defect site.
            perturb: The amount to perturb the sites in the supercell. Only perturb the sites with
                selective dynamics set to True. So this setting only works with `relax_radius`.
            target_frac_coords: If set, defect will be placed at the closest equivalent site to these
                fractional coordinates.
            return_site: If True, returns a tuple of the (defect supercell, defect site position).

        Returns:
            Structure: The supercell structure.
            PeriodicSite (optional): The position of the defect site in the supercell.
        """

        def _has_oxi(struct):
            return all([hasattr(site.specie, "oxi_state") for site in struct])

        if defect_structure is None:
            defect_structure = self.centered_defect_structure
        bulk_structure = center_structure(self.structure, self.site.frac_coords)
        keep_oxi = _has_oxi(bulk_structure) and _has_oxi(defect_structure)

        if sc_mat is None:
            sc_mat = get_sc_fromstruct(
                bulk_structure,
                min_atoms=min_atoms,
                max_atoms=max_atoms,
                min_length=min_length,
                force_diagonal=force_diagonal,
            )

        # Get the translation vector in to the desired cell in Cartesian coordinates
        if target_frac_coords is None:
            target_frac_coords = [1e-6, 1e-6, 1e-6]
        trans_R = np.floor(np.dot(target_frac_coords, sc_mat))
        trans_vec = np.dot(trans_R, bulk_structure.lattice.matrix)
        sc_defect_struct = bulk_structure * sc_mat

        # Use the site as a reference point for the position
        sc_mat_inv = np.linalg.inv(sc_mat)
        sc_pos = np.dot(self.site.frac_coords, sc_mat_inv)
        sc_site = PeriodicSite(
            species=self.site.specie,
            coords=sc_pos,
            lattice=sc_defect_struct.lattice,
            coords_are_cartesian=False,
        )

        bulk_site_mapping = _get_mapped_sites(bulk_structure, sc_defect_struct)
        defect_site_mapping = _get_mapped_sites(defect_structure, sc_defect_struct)

        # Remove the indices that that are not mapped
        rm_indices = set(bulk_site_mapping.values()) - set(defect_site_mapping.values())

        # Set the species for the sites that are mapped
        for defect_site, bulk_site in defect_site_mapping.items():
            sc_defect_struct[bulk_site]._species = defect_structure[defect_site].species

        # interstitials
        int_uc_indices = set(range(len(defect_structure))) - set(
            defect_site_mapping.keys()
        )
        for i in int_uc_indices:
            int_sc_pos = np.dot(defect_structure[i].frac_coords, sc_mat_inv)
            sc_defect_struct.insert(
                0,
                defect_structure[i].specie,
                int_sc_pos,
                coords_are_cartesian=False,
            )

        # Remove the sites that are not mapped
        sc_defect_struct.remove_sites(list(rm_indices))
        _set_selective_dynamics(
            structure=sc_defect_struct,
            site_pos=sc_site.coords,
            relax_radius=relax_radius,
        )
        if perturb is not None:
            _perturb_dynamic_sites(sc_defect_struct, distance=perturb)

        # Translate the structure to the target position
        sc_defect_struct.translate_sites(
            indices=list(range(len(sc_defect_struct))),
            vector=trans_vec,
            frac_coords=False,
        )
        sc_site._frac_coords += trans_R

        if not keep_oxi:
            sc_defect_struct.remove_oxidation_states()

        if dummy_species is not None:
            sc_defect_struct.insert(0, dummy_species, sc_site.frac_coords)

        if return_site:
            return sc_defect_struct, sc_site
        return sc_defect_struct

    @property
    def symmetrized_structure(self) -> SymmetrizedStructure:
        """Returns the multiplicity of a defect site within the structure.

        This is required for concentration analysis and confirms that defect_site is a
        site in bulk_structure.
        """
        sga = SpacegroupAnalyzer(
            self.structure, symprec=self.symprec, angle_tolerance=self.angle_tolerance
        )
        return sga.get_symmetrized_structure()

    def __eq__(self, __o: object) -> bool:
        """Equality operator."""
        if not isinstance(__o, Defect):
            raise TypeError("Can only compare Defects to Defects")
        sm = StructureMatcher(comparator=ElementComparator())
        return sm.fit(self.defect_structure, __o.defect_structure)

    @property
    def defect_type(self) -> int:
        """Get the defect type.

        Returns:
            int: The defect type.
        """
        return getattr(DefectType, self.__class__.__name__)

    @property
    def latex_name(self) -> str:
        """Get the latex name of the defect.

        Returns:
            str: The latex name of the defect.
        """
        root, suffix = self.name.split("_")
        return rf"{root}$_{{\rm {suffix}}}$"


class Vacancy(Defect):
    """Class representing a vacancy defect."""

    def get_multiplicity(self) -> int:
        """Returns the multiplicity of a defect site within the structure.

        This is required for concentration analysis and confirms that defect_site is
        a site in bulk_structure.

        Returns:
            int: The multiplicity of the defect.
        """
        symm_struct = self.symmetrized_structure
        defect_site = self.structure[self.defect_site_index]
        equivalent_sites = symm_struct.find_equivalent_sites(defect_site)
        return len(equivalent_sites)

    @property
    def name(self) -> str:
        """Name of the defect."""
        return f"v_{get_element(self.defect_site.specie)}"

    @property
    def defect_site(self):
        """Returns the site in the structure that corresponds to the defect site."""
        res = min(
            self.structure.get_sites_in_sphere(
                self.site.coords, 0.1, include_index=True
            ),
            key=lambda x: x[1],
        )
        return res

    @property
    def defect_site_index(self) -> int:
        """Get the index of the defect in the structure."""
        return self.defect_site.index

    @property
    def defect_structure(self):
        """Returns the defect structure with the proper oxidation state."""
        struct = self.structure.copy()
        struct.remove_sites([self.defect_site_index])
        return struct

    @property
    def element_changes(self) -> Dict[Element, int]:
        """Get the species changes of the vacancy defect.

        Returns:
            Dict[Element, int]: The species changes of the defect.
        """
        return {get_element(self.structure.sites[self.defect_site_index].specie): -1}

    def _guess_oxi_state(self) -> float:
        """Best guess for the oxidation state of the defect.

        For vacancies, the oxidation state is the opposite of the oxidation state of the
        removed atom.

        Returns:
            float: The oxidation state of the defect.
        """
        return -self.defect_site.specie.oxi_state

    def __repr__(self) -> str:
        """Representation of a vacancy defect."""
        vac_species = get_element(self.defect_site.specie)
        return f"{vac_species} Vacancy defect at site #{self.defect_site_index}"


class Substitution(Defect):
    """Single-site substitutional defects."""

    def __init__(
        self,
        structure: Structure,
        site: PeriodicSite,
        multiplicity: int | None = None,
        oxi_state: float | None = None,
        **kwargs,
    ) -> None:
        """Initialize a substitutional defect object.

        The position of `site` determines the atom to be removed and the species of
        `site` determines the replacing species.

        Args:
            structure: The structure of the defect.
            site: Replace the nearest site with this one.
            multiplicity: The multiplicity of the defect.
            oxi_state: The oxidation state of the defect, if not specified,
            this will be determined automatically.
        """
        super().__init__(structure, site, multiplicity, oxi_state, **kwargs)

    def get_multiplicity(self) -> int:
        """Returns the multiplicity of a defect site within the structure.

        This is required for concentration analysis and confirms that defect_site is
        a site in bulk_structure.
        """
        symm_struct = self.symmetrized_structure
        defect_site = self.structure[self.defect_site_index]
        equivalent_sites = symm_struct.find_equivalent_sites(defect_site)
        return len(equivalent_sites)

    @property
    def name(self) -> str:
        """Name of the defect."""
        return f"{get_element(self.site.specie)}_{get_element(self.defect_site.specie)}"

    @property
    def defect_structure(self) -> Structure:
        """Returns the defect structure."""
        struct: Structure = self.structure.copy()
        struct.remove_sites([self.defect_site_index])
        insert_el = get_element(self.site.specie)
        struct.insert(
            self.defect_site_index,
            species=insert_el,
            coords=np.mod(self.site.frac_coords, 1),
        )
        return struct

    @property
    def defect_site(self):
        """Returns the site in the structure that corresponds to the defect site."""
        return min(
            self.structure.get_sites_in_sphere(
                self.site.coords, 0.1, include_index=True
            ),
            key=lambda x: x[1],
        )

    @property
    def defect_site_index(self) -> int:
        """Get the index of the defect in the structure."""
        return self.defect_site.index

    @property
    def element_changes(self) -> Dict[Element, int]:
        """Get the species changes of the substitution defect.

        Returns:
            Dict[Element, int]: The species changes of the defect.
        """
        return {
            get_element(self.structure.sites[self.defect_site_index].specie): -1,
            get_element(self.site.specie): +1,
        }

    def _guess_oxi_state(self) -> float:
        """Best guess for the oxidation state of the defect.

        For a substitution defect, the oxidation state of the defect is given
        by the difference between the oxidation state of the new and old atoms.

        Returns:
            float: The oxidation state of the defect.
        """
        rm_oxi = self.structure[self.defect_site_index].specie.oxi_state

        # check if substitution atom is present in structure (i.e. antisite substitution):
        sub_elt_sites_in_struct = [
            site
            for site in self.structure
            if site.specie.symbol
            == self.site.specie.symbol  # gives Element symbol (without oxi state)
        ]
        if len(sub_elt_sites_in_struct) == 0:
            sub_states = self.site.specie.common_oxidation_states
            if len(sub_states) == 0:
                raise ValueError(
                    f"No common oxidation states found for {self.site.specie}."
                    "Please specify the oxidation state manually."
                )
            sub_oxi = min(sub_states, key=lambda x: abs(x - rm_oxi))
        else:
            sub_oxi = int(
                np.mean([site.specie.oxi_state for site in sub_elt_sites_in_struct])
            )

        return sub_oxi - rm_oxi

    def __repr__(self) -> str:
        """Representation of a substitutional defect."""
        rm_species = get_element(self.defect_site.specie)
        sub_species = get_element(self.site.specie)
        return (
            f"{sub_species} subsitituted on the {rm_species} site at "
            f"at site #{self.defect_site_index}"
        )


class Interstitial(Defect):
    """Interstitial Defect."""

    def __init__(
        self,
        structure: Structure,
        site: PeriodicSite,
        multiplicity: int = 1,
        oxi_state: float | None = None,
        equivalent_sites: list[PeriodicSite] | None = None,
        **kwargs,
    ) -> None:
        """Initialize an interstitial defect object.

        The interstitial defect effectively inserts the `site` object into the structure.

        Args:
            structure: The structure of the defect.
            site: Inserted site, also determines the species.
            multiplicity: The multiplicity of the defect.
            oxi_state: The oxidation state of the defect, if not specified,
                this will be determined automatically.
        """
        super().__init__(
            structure, site, multiplicity, oxi_state, equivalent_sites, **kwargs
        )

    def get_multiplicity(self) -> int:
        """Determine the multiplicity of the defect site within the structure."""
        raise NotImplementedError(
            "Interstitial multiplicity should be determined by the generator."
        )

    @property
    def name(self) -> str:
        """Name of the defect."""
        return f"{get_element(self.site.specie)}_i"

    @property
    def defect_structure(self) -> Structure:
        """Returns the defect structure."""
        struct: Structure = self.structure.copy()
        # use the highest value oxidation state among the two most popular ones
        # found in the ICSD
        inter_states = self.site.specie.icsd_oxidation_states[:2]
        if len(inter_states) == 0:
            _logger.warning(
                f"No oxidation states found for {self.site.specie.symbol}. "
                "in ICSD using `oxidation_states` without frequency ranking."
            )
            inter_states = self.site.specie.oxidation_states
        inter_oxi = max(inter_states, key=abs)
        int_specie = Species(self.site.specie.symbol, inter_oxi)
        struct.insert(
            0,
            species=int_specie,
            coords=np.mod(self.site.frac_coords, 1),
        )
        return struct

    @property
    def defect_site_index(self) -> int:
        """Get the index of the defect in the structure."""
        return 0

    @property
    def element_changes(self) -> Dict[Element, int]:
        """Get the species changes of the intersitial defect.

        Returns:
            Dict[Element, int]: The species changes of the defect.
        """
        return {
            get_element(self.site.specie): +1,
        }

    def _guess_oxi_state(self) -> float:
        """Best guess for the oxidation state of the defect.

        For interstitials, just use the oxidation state of the site.
        The oxidation of the interstitial site is determined by highest
        absolute value of the oxidation states of the inserted atom.

        Returns:
            float: The oxidation state of the defect.
        """
        sub_site = self.defect_structure[self.defect_site_index]
        return sub_site.specie.oxi_state

    def __repr__(self) -> str:
        """Representation of a interstitial defect."""
        sub_species = get_element(self.site.specie)
        fpos_str = ",".join(f"{x:.2f}" for x in self.site.frac_coords)
        return f"{sub_species} intersitial site at [{fpos_str}]"


class DefectComplex(Defect):
    """A complex of defects."""

    def __init__(
        self,
        defects: list[Defect],
        oxi_state: float | None = None,
    ) -> None:
        """Initialize a complex defect object.

        Args:
            defects: List of defects.
            oxi_state: The oxidation state of the defect, if not specified,
                this will be determined automatically.
        """
        self.defects = defects
        self.structure = self.defects[0].structure
        self.oxi_state = self._guess_oxi_state() if oxi_state is None else oxi_state
        defect_sites = [d.site for d in self.defects]
        center_of_mass = np.mean([s.coords for s in defect_sites], axis=0)
        self.site = PeriodicSite(
            species=DummySpecies(),
            coords=center_of_mass,
            lattice=self.structure.lattice,
            coords_are_cartesian=True,
        )

    def __repr__(self) -> str:
        """Representation of a complex defect."""
        return f"Complex defect containing: {[d.name for d in self.defects]}"

    def __eq__(self, __o: object) -> bool:
        """Check if  are equal."""
        if not isinstance(__o, Defect):
            raise TypeError("Can only compare Defects to Defects")
        sm = StructureMatcher(comparator=ElementComparator())
        this_structure = self.defect_structure_with_com
        if isinstance(__o, DefectComplex):
            that_structure = __o.defect_structure_with_com
        else:
            that_structure = __o.defect_structure
        return sm.fit(this_structure, that_structure)

    @property
    def defect_structure_with_com(self) -> Structure:
        """Returns the defect structure with the center of mass as dummy site."""
        struct = self.defect_structure.copy()
        struct.insert(0, self.site.specie, self.site.frac_coords)
        return struct

    def get_multiplicity(self) -> int:
        """Determine the multiplicity of the defect site within the structure."""
        raise NotImplementedError("Not implemented for defect complexes")

    @property
    def element_changes(self) -> Dict[Element, int]:
        """Determine the species changes of the complex defect."""
        cnt: dict[Element, int] = collections.defaultdict(int)
        for defect in self.defects:
            for el, change in defect.element_changes.items():
                cnt[el] += change
        return dict(cnt)

    @property
    def name(self) -> str:
        """Name of the defect."""
        return "+".join([d.name for d in self.defects])

    def _guess_oxi_state(self) -> float:
        oxi_state = 0.0
        for defect in self.defects:
            oxi_state += defect.oxi_state
        return oxi_state

    @property
    def defect_structure(self) -> Structure:
        """Returns the defect structure."""
        defect_structure = self.structure.copy()
        for defect in self.defects:
            update_structure(
                defect_structure, defect.site, defect_type=defect.defect_type
            )
        return defect_structure


def update_structure(structure, site, defect_type):
    """Update the structure with the defect site.

    Types of operations:
        1. Vacancy: remove the site.
        2. Substitution: replace the site with the defect species.
        3. Interstitial: insert the defect species at the site.

    Args:
        structure: The structure to be updated.
        site: The defect site.
        defect_type: The type of the defect.

    Returns:
        Structure: The updated structure.
    """

    def _update(structure, site, rm: bool, replace: bool):
        in_sphere = structure.get_sites_in_sphere(site.coords, 0.1, include_index=True)

        if len(in_sphere) == 0 and rm:  # pragma: no cover
            raise ValueError("No site found to remove.")

        if rm or replace:
            rm_site = min(
                in_sphere,
                key=lambda x: x[1],
            )
            rm_index = rm_site.index
            structure.remove_sites([rm_index])

        if rm:
            return

        sub_specie = Element(site.specie.symbol)
        structure.insert(
            0,
            species=sub_specie,
            coords=site.frac_coords,
        )

    if defect_type == DefectType.Vacancy:
        _update(structure, site, rm=True, replace=False)
    elif defect_type == DefectType.Substitution:
        _update(structure, site, rm=False, replace=True)
    elif defect_type == DefectType.Interstitial:
        _update(structure, site, rm=False, replace=False)
    else:
        raise ValueError("Unknown point defect type.")


class Adsorbate(Interstitial):
    """Subclass of Interstitial with a different name.

    Used for keeping track of adsorbate, which are treated the same
    algorithmically as interstitials, but are conceptually separate.
    """

    @property
    def name(self) -> str:
        """Returns a name for this defect."""
        return f"{get_element(self.site.specie)}_{{ads}}"

    def __repr__(self) -> str:
        """Representation of a adsorbate defect."""
        sub_species = get_element(self.site.specie)
        fpos_str = ",".join(f"{x:.2f}" for x in self.site.frac_coords)
        return f"{sub_species} adsorbate site at [{fpos_str}]"


def get_element(sp_el: Species | Element) -> Element:
    """Get the element from a species or element."""
    if isinstance(sp_el, Species):
        return sp_el.element
    return sp_el


def get_vacancy(structure: Structure, isite: int, **kwargs) -> Vacancy:
    """Get a vacancy defect from a structure and site index.

    Convenience function for creating a Vacancy object quickly.

    Args:
        structure: The structure to create the vacancy in.
        isite: The site index of the vacancy.
        **kwargs: Keyword arguments to pass to Vacancy constructor.
    """
    site = structure[isite]
    return Vacancy(structure=structure, site=site, **kwargs)


def _set_selective_dynamics(structure, site_pos, relax_radius):
    if relax_radius is None:
        return
    if relax_radius == "auto":
        relax_radius = min(get_plane_spacing(structure.lattice.matrix)) / 2.0
    if not isinstance(relax_radius, float):
        raise ValueError("relax_radius must be a float or 'auto' or None")
    structure.get_sites_in_sphere(site_pos, relax_radius)
    relax_sites = structure.get_sites_in_sphere(
        site_pos, relax_radius, include_index=True
    )
    relax_indices = [site.index for site in relax_sites]
    relax_mask = [[False, False, False]] * len(structure)
    for i in relax_indices:
        relax_mask[i] = [True, True, True]
    structure.add_site_property("selective_dynamics", relax_mask)


def perturb_sites(
    structure,
    distance: float,
    min_distance: float | None = None,
    site_indices: list | None = None,
) -> None:
    """
    Performs a random perturbation of the sites in a structure to break
    symmetries.

    Args:
        structure (Structure): Input structure.
        distance (float): Distance in angstroms by which to perturb each
            site.
        min_distance (None, int, or float): if None, all displacements will
            be equal amplitude. If int or float, perturb each site a
            distance drawn from the uniform distribution between
            'min_distance' and 'distance'.
        site_indices (list): List of site indices on which to perform the
            perturbation. If None, all sites will be perturbed.

    """

    def get_rand_vec():
        # deals with zero vectors.
        vector = np.random.randn(3)
        vnorm = np.linalg.norm(vector)
        dist = distance
        if isinstance(min_distance, (float, int)):
            dist = np.random.uniform(min_distance, dist)
        return vector / vnorm * dist if vnorm != 0 else get_rand_vec()

    if site_indices is None:
        site_indices_ = list(range(len(structure._sites)))
    else:
        site_indices_ = site_indices

    for i in site_indices_:
        structure.translate_sites([i], get_rand_vec(), frac_coords=False)


def _perturb_dynamic_sites(structure, distance):
    free_indices = [
        i
        for i, site in enumerate(structure)
        if site.properties["selective_dynamics"][0]
    ]
    perturb_sites(structure=structure, distance=distance, site_indices=free_indices)


def _get_mapped_sites(uc_structure: Structure, sc_structure: Structure, r=0.001):
    """Get the list of sites indices in the supercell corresponding to the unit cell."""
    mapped_site_indices = {}
    for isite, uc_site in enumerate(uc_structure):
        sc_sites = sc_structure.get_sites_in_sphere(uc_site.coords, r)
        if len(sc_sites) == 1:
            mapped_site_indices[isite] = sc_sites[0].index
    return mapped_site_indices


def center_structure(structure, ref_fpos) -> Structure:
    """Shift the sites around a center.

    Move all the sites in the structure so that they
    are in the periodic image closest to the reference fractional position.
    """
    struct = structure.copy()
    for idx, d_site in enumerate(struct):
        _, jiimage = struct.lattice.get_distance_and_image(ref_fpos, d_site.frac_coords)
        struct.translate_sites([idx], jiimage, to_unit_cell=False)
    return struct
